@@ -28,6 +28,9 @@ class Kattis(requests.Session):
         def __init__(self, data):
             super().__init__(data)
             self.to_df = lambda: pd.DataFrame(data)
+    
+    def set_base_url(self, url):
+        self.BASE_URL = url
 
     def __init__(self, user, password=None):
         '''
@@ -52,7 +55,7 @@ class Kattis(requests.Session):
             'password': self.password
         }
         response = self.post(f'{self.BASE_URL}/login/email', data=data)
-        assert response.url == 'https://open.kattis.com/', 'Cannot login to Kattis'
+        assert response.url.startswith(self.BASE_URL), 'Cannot login to Kattis'
         print('Logged in to Kattis!', flush=True)
 
         self.homepage = bs(response.content, features='lxml')
@@ -74,7 +77,7 @@ class Kattis(requests.Session):
         '''
         Gets all Kattis problems based on some filters.
         Returns a JSON-like structure with these fields:
-            name, fastest, shortest, total, acc, difficulty, category, link
+            name, id, fastest, shortest, total, acc, difficulty, category, link
 
         Default: all solved and partially solved problems.
         '''
@@ -100,9 +103,12 @@ class Kattis(requests.Session):
                 for f in as_completed(futures):
                     response = f.result()
                     soup = bs(response.content, features='lxml')
+                    if not soup: continue
                     table = soup.find('section', class_='strip strip-item-plain').find('table', class_='table2')
-                    try: table_content = table.tbody.find_all('tr')
-                    except AttributeError: continue
+                    try:
+                        table_content = table.tbody.find_all('tr')
+                    except AttributeError:
+                        continue
                     for row in table_content:
                         columns = row.find_all('td')
                         if columns:
@@ -121,10 +127,16 @@ class Kattis(requests.Session):
                                     # - difficulty 5.0 -> [5.0]
                             except:
                                 difficulty = None
-                            try: category = re.findall('[A-Za-z]+', columns[7].text)[0]
-                            except: category = 'N/A'
+                            if len(columns) == 10:
+                                try:
+                                    category = re.findall('[A-Za-z]+', columns[7].text)[0]
+                                except:
+                                    category = 'N/A'
+                            else:
+                                category = 'N/A'
                             data.append({
                                 'name': name,
+                                'id': link.split('/')[-1],
                                 'fastest': fastest,
                                 'shortest': shortest,
                                 'total': total,
@@ -133,7 +145,7 @@ class Kattis(requests.Session):
                                 'category': category,
                                 'link': link
                             })
-        return self.Result(data)
+        return self.Result(sorted(data, key=lambda x: x['id']))
 
     @lru_cache
     def plot_problems(self, filepath=None, show_solved=True, show_partial=True, show_tried=False, show_untried=False):
@@ -152,7 +164,7 @@ class Kattis(requests.Session):
         hist = sns.histplot(data=df, x='difficulty', hue='category', multiple='stack', binwidth=0.1, hue_order=hue_order, palette=palette)
         hist.set(title=f'Solved Kattis Problems by {self.user} ({df.shape[0]})', xlabel='Difficulty')
         plt.legend(title='Category', loc='upper right', labels=hue_order[::-1])
-        plt.xticks([*range(math.floor(min(df.difficulty)), math.ceil(max(df.difficulty))+1)])
+        plt.xticks([*range(math.floor(min([d for d in df.difficulty if d], default=0)), math.ceil(max([d for d in df.difficulty if d], default=0))+1)])
         if filepath != None:
             plt.savefig(filepath)
             print(f'Saved to {filepath}')
@@ -182,37 +194,57 @@ class Kattis(requests.Session):
 
         soup = bs(response.content, features='lxml')
         body = soup.find('div', class_='problembody')
-        data = {'id': problem_id, 'text': body.text.strip(), 'metadata': {}}
-        meta = data['metadata']
+        data = {'id': problem_id, 'text': body.text.strip()}
 
-        for div in soup.find_all('div', class_='metadata_list-item'):
-            div_text = div.text.strip().split('\n')
-            if div_text[0] == 'CPU Time limit':
-                meta['cpu'] = div_text[-1].strip()
-            elif div_text[0] == 'Memory limit':
-                meta['memory'] = div_text[-1].strip()
-            elif div_text[0] == 'Difficulty':
-                meta['difficulty'] = float(re.findall('[\d\.]+', div_text[-1])[-1])
-                meta['category'] = re.findall('[A-Za-z]+', div_text[-1])[0]
-            elif div_text[0] == 'Author':
-                meta['author'] = div_text[-1].strip()
-            elif div_text[0] == 'Source':
-                meta['source'] = div_text[-1].strip()
-            elif div_text[0] == 'Attachments' or div_text[0] == 'Downloads':
-                meta['files'] = meta.get('files', {})
-                for url, fn in [(f"{self.BASE_URL}{a.get('href')}", a.get('download') or a.get('href').split('/')[-1]) for a in div.find_all('a')]:
-                    if url.endswith('zip'):
-                        with zipfile.ZipFile(io.BytesIO(self.get(url).content)) as z:
-                            meta['files'][fn] = {}
-                            for inner_fn in z.namelist():
-                                with z.open(inner_fn) as inner_file:
-                                    meta['files'][fn][inner_fn] = inner_file.read().decode("utf-8")
-                    else:
-                        meta['files'][fn] = self.get(url).text
+        cpu = memory = author = source = ''
+        difficulty = None
+        category = 'N/A'
+        files = {}
+        for div in soup.find_all('div', class_='metadata-grid'):
+            for d in div.find_all('div', class_='card'):
+                div_text = [s.text.strip() for s in d.find_all('span') if s.text.strip()]
+                if div_text[0] == 'CPU Time limit':
+                    cpu = div_text[-1].strip()
+                elif div_text[0] == 'Memory limit':
+                    memory = div_text[-1].strip()
+                elif len(div_text) > 1 and div_text[1] == 'Difficulty':
+                    try:
+                        difficulty = float(div_text[0])
+                    except:
+                        difficulty = None
+                    try:
+                        category = div_text[2].strip()
+                    except:
+                        category = 'N/A'
+                elif div_text[0] == 'Source & License':
+                    text_links = [(s.text.strip(), [a.get('href').strip('/') for a in s.find_all('a')]) for s in d.find_all('span') if s.text.strip()]
+                    for text, links in text_links:
+                        if any(link.startswith('problem-authors') for link in links): author = text
+                        if any(link.startswith('problem-sources') for link in links): source = text
+                elif div_text[0] == 'Attachments' or div_text[0] == 'Downloads':
+                    for url, fn in [(f"{self.BASE_URL}{a.get('href')}", a.get('download') or a.get('href').split('/')[-1]) for a in d.find_all('a')]:
+                        if url.endswith('zip'):
+                            with zipfile.ZipFile(io.BytesIO(self.get(url).content)) as z:
+                                files[fn] = {}
+                                for inner_fn in z.namelist():
+                                    with z.open(inner_fn) as inner_file:
+                                        files[fn][inner_fn] = inner_file.read().decode("utf-8")
+                        else:
+                            files[fn] = self.get(url).text
+        data = {
+            **data,
+            'cpu': cpu,
+            'memory': memory,
+            'difficulty': difficulty,
+            'category': category,
+            'author': author,
+            'source': source,
+            'files': files
+        }
 
         # statistics
         response = self.get(f'{self.BASE_URL}/problems/{problem_id}/statistics')
-        meta['statistics'] = {}
+        data['statistics'] = {}
         soup = bs(response.content, features='lxml')
         category_map = {}
         for option in soup.find_all('option'):
@@ -221,9 +253,9 @@ class Kattis(requests.Session):
             table = section.find('table', class_='table2 report_grid-problems_table')
             section_id = section.get('id')
             language, description = category_map[section_id]
-            meta['statistics'][language] = meta['statistics'].get(language, {})
-            meta['statistics'][language][['fastest', 'shortest']['shortest' in section_id]] = {}
-            stats = meta['statistics'][language][['fastest', 'shortest']['shortest' in section_id]]
+            data['statistics'][language] = data['statistics'].get(language, {})
+            data['statistics'][language][['fastest', 'shortest']['shortest' in section_id]] = {}
+            stats = data['statistics'][language][['fastest', 'shortest']['shortest' in section_id]]
             if table:
                 stats['ranklist'] = []
                 for row in table.tbody.find_all('tr'):
@@ -247,25 +279,30 @@ class Kattis(requests.Session):
 
         # my submissions
         response = self.get(f'{self.BASE_URL}/problems/{problem_id}?tab=submissions')
-        meta['submissions'] = []
+        data['submissions'] = []
         soup = bs(response.content, features='lxml')
         table = soup.find('table', id='submissions')
-        for row in table.tbody.find_all('tr'):
-            columns = row.find_all('td')
-            columns_text = [column.text for column in columns if column.text]
-            if columns_text:
-                status, runtime, language, tc, _ = columns_text
-                link = f"{self.BASE_URL}{columns[-1].find('a').get('href')}"
-                runtime = ' '.join(runtime.split())
-                test_case_passed, test_case_full = map(int, tc.split('/'))
-                meta['submissions'].append({
-                    'status': status,
-                    'runtime': runtime,
-                    'language': language,
-                    'test_case_passed': test_case_passed,
-                    'test_case_full': test_case_full,
-                    'link': link
-                })
+        if table:
+            for row in table.tbody.find_all('tr'):
+                columns = row.find_all('td')
+                columns_text = [column.text.strip() for column in columns if column.text.strip()]
+                if columns_text:
+                    try:
+                        status, runtime, language, tc, *_ = columns_text
+                        runtime = ' '.join(runtime.split())
+                        test_case_passed, test_case_full = map(int, tc.split('/'))
+                    except:
+                        status, language, *_ = columns_text
+                        runtime = test_case_passed = test_case_full = None
+                    link = f"{self.BASE_URL}{columns[-1].find('a').get('href')}"
+                    data['submissions'].append({
+                        'status': status,
+                        'runtime': runtime,
+                        'language': language,
+                        'test_case_passed': test_case_passed,
+                        'test_case_full': test_case_full,
+                        'link': link
+                    })
 
         ret = [data]
         for pid in set(problem_ids) - {problem_id}:
@@ -273,6 +310,80 @@ class Kattis(requests.Session):
             if prob: ret.append(prob[0])
         return self.Result(ret)
 
+    @lru_cache
+    def achievements(self, verbose=False):
+        '''
+        Lists down all your Kattis achievements. Flex it!
+        The parameter verbose adjusts whether you only want to include the problems
+            with at least one achievement or not.
+
+        Default: includes only solved questions with at least one achievement.
+        '''
+
+        has_content = True
+        params = {'page': 1}
+        data = []
+        with ThreadPoolExecutor(max_workers=self.MAX_WORKERS) as executor:
+            futures = []
+            while has_content:
+                has_content = False
+                futures.clear()
+                for _ in range(self.MAX_WORKERS):
+                    futures.append(executor.submit(self.get, f'{self.BASE_URL}/users/{self.user}', params=params.copy()))
+                    params['page'] += 1
+                for f in as_completed(futures):
+                    response = f.result()
+                    soup = bs(response.content, features='lxml')
+                    table = soup.find('table', class_='table2')
+                    for row in table.tbody.find_all('tr'):
+                        columns = row.find_all('td')
+                        columns_text = [column.text for column in columns if column.text]
+                        if columns_text:
+                            if columns[0].find('a') == None: continue
+                            has_content = True
+                            link = f"{self.BASE_URL}{columns[0].find('a').get('href')}"
+                            name = columns[0].text
+                            runtime = float(columns[1].text.replace('--', 'inf'))
+                            length = int(columns[2].text.replace('--', '-1'))
+                            if len(columns) == 3:
+                                if not verbose: continue
+                                achievement = ''
+                                difficulty = None
+                                category = 'N/A'
+                            elif len(columns) == 4:
+                                if not verbose: continue
+                                achievement = ''
+                                try:
+                                    difficulty = float(re.findall('[\d\.]+', columns[3].text)[-1])
+                                except:
+                                    difficulty = None
+                                try:
+                                    category = re.findall('[A-Za-z]+', columns[3].text)[0]
+                                except:
+                                    category = 'N/A'
+                            else:
+                                achievement = ', '.join(sorted(set(sp.text.strip() for sp in columns[3].find_all('span') if len(sp.find_all('span')) == 1)))
+                                if not achievement: continue
+                                try:
+                                    difficulty = float(re.findall('[\d\.]+', columns[4].text)[-1])
+                                except:
+                                    difficulty = None
+                                try:
+                                    category = re.findall('[A-Za-z]+', columns[4].text)[0]
+                                except:
+                                    category = 'N/A'
+                            data.append({
+                                'name': name,
+                                'id': link.split('/')[-1],
+                                'runtime': runtime,
+                                'length': length,
+                                'achievement': achievement,
+                                'difficulty': difficulty,
+                                'category': category,
+                                'link': link
+                            })
+        return self.Result(sorted(data, key=lambda x: x['id']))
+
     @lru_cache
     def stats(self, language='', *languages):
         '''
@@ -309,7 +420,7 @@ class Kattis(requests.Session):
                     table = soup.find('table', class_='table2 report_grid-problems_table double-rows')
                     for row in table.tbody.find_all('tr'):
                         columns = row.find_all('td')
-                        columns_text = [column.text for column in columns if column.text]
+                        columns_text = [column.text.strip() for column in columns if column.text.strip()]
                         if columns_text:
                             has_content = True
                             link = f"{self.BASE_URL}/submissions/{columns[-1].find('a').get('href').split('/')[-1]}"
@@ -336,12 +447,12 @@ class Kattis(requests.Session):
                             else:
                                 data[pid] = max(
                                     data[pid], new_data,
-                                    key=lambda x: (x.get('score'), x['test_case_passed'], -float(x['runtime'] if '>' not in x['runtime'] else 1e9))
+                                    key=lambda x: (x.get('score', tc_pass/tc_full), x['test_case_passed'], -float(x['runtime'] if '>' not in x['runtime'] else 1e9))
                                 )
         ret = [{'id': k, **v} for k, v in data.items()]
         for lang in set(languages) - {language}:
             ret.extend(self.stats(lang))
-        return self.Result(ret)
+        return self.Result(sorted(ret, key=lambda x: x['id']))
 
     @lru_cache
     def suggest(self):
@@ -351,7 +462,10 @@ class Kattis(requests.Session):
         '''
 
         soup = self.homepage
-        table = soup.find_all('table', class_='table2 report_grid-problems_table')[0]
+        try:
+            table = soup.find_all('table', class_='table2 report_grid-problems_table')[0]
+        except:
+            return self.Result([])
         data = []
         for row in table.tbody.find_all('tr'):
             header = row.find('th')
@@ -381,7 +495,12 @@ class Kattis(requests.Session):
 
         if country == university == None:
             soup = self.homepage
-            table = soup.find_all('table', class_='table2 report_grid-problems_table')[1]
+            try:
+                table = soup.find_all('table', class_='table2 report_grid-problems_table')[1]
+            except:
+                return self.Result([])
+            if not table:
+                return self.Result([])
             data = []
             for row in table.tbody.find_all('tr'):
                 columns = row.find_all('td')
@@ -413,8 +532,12 @@ class Kattis(requests.Session):
             country_code = guess_id(country, COUNTRIES)
             response = self.get(f'{self.BASE_URL}/countries/{country_code}')
             soup = bs(response.content, features='lxml')
-            table = soup.find('table', class_='table2 report_grid-problems_table', id='top_users')
-            if not table: return []
+            try:
+                table = soup.find('table', class_='table2 report_grid-problems_table', id='top_users')
+            except:
+                return self.Result([])
+            if not table:
+                return self.Result([])
             data = []
             headers = [re.findall(r'[A-Za-z]+', h.text)[0] for h in table.find_all('th')]
             for row in table.tbody.find_all('tr'):
@@ -459,7 +582,8 @@ class Kattis(requests.Session):
             response = self.get(f'{self.BASE_URL}/universities/{university_code}')
             soup = bs(response.content, features='lxml')
             table = soup.find('table', class_='table2 report_grid-problems_table', id='top_users')
-            if not table: return []
+            if not table:
+                return self.Result([])
             data = []
             headers = [re.findall(r'[A-Za-z]+', h.text)[0] for h in table.find_all('th')]
             for row in table.tbody.find_all('tr'):
@@ -500,3 +624,89 @@ class Kattis(requests.Session):
                     'university': UNIVERSITIES[university_code]
                 })
         return self.Result(data)
+
+    @lru_cache
+    def problem_authors(self):
+        '''
+        Lists down all problem authors alongside with the number of problems one has
+            authored as well as the average difficulty.
+        '''
+
+        response = self.get(f'{self.BASE_URL}/problem-authors')
+        try:
+            soup = bs(response.content, features='lxml')
+        except:
+            return self.Result([])
+        table = soup.find('table', class_='table2')
+        if not table:
+            return self.Result([])
+        data = []
+        for row in table.tbody.find_all('tr'):
+            columns = row.find_all('td')
+            columns_text = [column.text.strip() for column in columns]
+            columns_url = [column.find_all('a') for column in columns]
+            name = columns_text[0].strip()
+            link = f'{self.BASE_URL}{columns_url[0][0].get("href")}'
+            num_problems = int(columns_text[1])
+            try:
+                difficulty = float(re.findall('[\d\.]+', columns_text[2])[-1])
+            except:
+                difficulty = None
+            try:
+                category = re.findall('[A-Za-z]+', columns_text[2])[0]
+            except:
+                category = 'N/A'
+            data.append({
+                'name': name,
+                'problems': num_problems,
+                'avg_difficulty': difficulty,
+                'avg_category': category,
+                'link': link
+            })
+        return self.Result(data)
+    
+    @lru_cache
+    def problem_sources(self):
+        '''
+        Lists down all problem sources alongside with the number of problems one comes from
+            as well as the average difficulty.
+        '''
+
+        response = self.get(f'{self.BASE_URL}/problem-sources')
+        try:
+            soup = bs(response.content, features='lxml')
+        except:
+            return self.Result([])
+        table = soup.find('table', class_='table2')
+        if not table:
+            return self.Result([])
+        data = []
+        for row in table.tbody.find_all('tr'):
+            columns = row.find_all('td')
+            columns_text = [column.text.strip() for column in columns]
+            columns_url = [column.find_all('a') for column in columns]
+            name = columns_text[0].strip()
+            link = f'{self.BASE_URL}{columns_url[0][0].get("href")}'
+            num_problems = int(columns_text[1])
+            try:
+                difficulty = float(re.findall('[\d\.]+', columns_text[2])[-1])
+            except:
+                difficulty = None
+            try:
+                category = re.findall('[A-Za-z]+', columns_text[2])[0]
+            except:
+                category = 'N/A'
+            data.append({
+                'name': name,
+                'problems': num_problems,
+                'avg_difficulty': difficulty,
+                'avg_category': category,
+                'link': link
+            })
+        return self.Result(data)
+
+class NUSKattis(Kattis):
+    def __init__(self, user, password=None):
+        print('Logging in to NUS Kattis...', flush=True)
+        self.set_base_url('https://nus.kattis.com')
+        super().__init__(user, password)
